#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <getopt.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ncurses.h>
#include <alsaplayer/control.h>

#define USAGE		"usage: gruke <directory>\n\nEnter '?' when gruke is running for a list of keystrokes.\n"
#define MAX_DIR_NO	2500
#define MAX_NAME_LEN	128
#define MAX_PATH_LEN	256
#define MAX_LIST_LEN	2500
#define TOP_SPACE	2
#define LEFT_SPACE	5
#define PAGE_LEN	15
#define MAX_SAVE_LEN	32
#define SAVE_DIR	"/jukebox/playlists"
#define CFG		".cfg"
#define MAX_DEPTH	5

typedef struct song song_t;

struct song
{
	char name[MAX_NAME_LEN];
	char path[MAX_PATH_LEN];
	int start;
	int end;
	int isdir;
};

typedef struct depth depth_t;

struct depth
{
	int startpos;
	int selected;
};

song_t dir_list[MAX_DIR_NO];
song_t play_list[MAX_LIST_LEN];
song_t current;

int selected=0;
int playing=0;
int looping=0;
char *validchars="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789 '";

int check_validchars(char c)
{
	int i=0;

	for(i=0; i<strlen(validchars); i++)
		if(c==validchars[i]) return 1;
	return 0;
}

char *read_next_line(FILE *fp)
{
	int c;
	int i=0;
	char *retval=NULL;
	while((c=fgetc(fp))!=EOF)
	{
		retval=realloc(retval, i+2);
		retval[i++]=c;
		if(c=='\n') break;
	}
	if(i) retval[i]=0;
	return retval;
}

int check_digits(char *string)
{
	char *cp=NULL;
	for(cp=string; *cp; cp++)
	{
		if(!(isdigit(*cp)))
		{
			endwin();
			fprintf(stderr, "Non-numeric value in %s file timer settings.\n", CFG);
			return -1;
		}
	}
	return 0;
}

int read_config_file(const char *config, const char *cur_dir, int *count)
{
	FILE *fp=NULL;
	char *data=NULL;

	if(!(fp=fopen(config, "r")))
	{
		endwin();
		fprintf(stderr, "Could not open %s file: %s\n", CFG, strerror(errno));
		return -1;
	}
	while((data=read_next_line(fp)))
	{
		char *start=NULL;
		char *end=NULL;
		char *name=NULL;
		char *np=NULL;
		if((np=strrchr(data, '\n'))) *np='\0';
		if(	   !(start=strchr(data, ':'))
			|| !(end=strchr(start+1, ':'))
			|| !(name=strchr(end+1, ':')))
		{
			endwin();
			fprintf(stderr, "Could not parse %s file. Line:'%s'\n", CFG, data);
			free(data);
			return -1;
		}
		*start='\0'; start++;
		*end='\0'; end++;
		*name='\0'; name++;
		if(check_digits(start) || check_digits(end))
		{
			free(data);
			return -1;
		}
		snprintf(dir_list[*count].name, MAX_NAME_LEN, name);

		/* saved playlists already have the full path... */
		if(data[0]=='/')
			snprintf(dir_list[*count].path, MAX_PATH_LEN, data);
		else
			snprintf(dir_list[*count].path, MAX_PATH_LEN, "%s/%s",
							cur_dir, data);
		dir_list[*count].start=atoi(start);
		dir_list[*count].end=atoi(end);
		dir_list[*count].isdir=0;
		(*count)++;
		free(data);
	}
	if(fclose(fp))
	{
		endwin();
		fprintf(stderr, "Closing %s file: %s\n", strerror(errno), CFG);
		return -1;
	}
	
	return 0;
}

int get_fullpath(const char *id, char *dest, int len)
{
	char cwd[1024]="";

	if(id[0]!='/')
	{
		if(!getcwd(cwd, sizeof(cwd)))
		{
			fprintf(stderr, "Could not get current working directory\n");
			return -1;
		}
		snprintf(dest, len, "%s%s%s", cwd, strcmp(cwd, "/")==0?"":"/", id);
	}

	return 0;
}

int alloc_new_dir(const char *name, int arr_ind)
{
	snprintf(dir_list[arr_ind].name, sizeof(dir_list[arr_ind].name), name);
	if(get_fullpath(name, dir_list[arr_ind].path, MAX_PATH_LEN))
	{
		fprintf(stderr, "Unable to get full path of song '%s'.\n", name);
		return -1;
	}
	dir_list[arr_ind].isdir=1;
	return 0;
}

int get_dir(const char *dir_to_scan, int *count)
{
	int i=0;
	int n=-1;
	struct dirent **dir;
	struct stat buf;

	if((n=scandir(dir_to_scan, &dir, 0, alphasort))<0) perror("scandir");
	for(i=0; i<n; i++)
	{
		if(dir[i]->d_ino==0) { free(dir[i]); continue; }
		if((strcmp(dir[i]->d_name, "."))==0
			|| (strcmp(dir[i]->d_name, ".."))==0) { free(dir[i]); continue; }

		if(lstat(dir[i]->d_name, &buf))
		{
			endwin();
			fprintf(stderr, "%s: Unable to stat: %s.\n", dir[i]->d_name, strerror(errno));
			return -1;
		}
		
		if(S_ISDIR(buf.st_mode))
		{
			if((alloc_new_dir(dir[i]->d_name, *count))) return -1;
			(*count)++;
		}
		if(strcmp(dir[i]->d_name, CFG)==0)
		{
			if(read_config_file(dir[i]->d_name, dir_to_scan, count))
				return -1;
		}
		free(dir[i]);
	}

	return 0;
}

int print_truncated(char *string, int row, int col, int sel)
{
	int k=LEFT_SPACE;
	char *cp=NULL;

	if(sel) mvprintw(row, LEFT_SPACE-1, "*");
	else mvprintw(row, LEFT_SPACE-1, " ");

	for(cp=string; (*cp && k<col); cp++)
				mvprintw(row, k++, "%c", *cp);
	while(k<col)
		mvprintw(row, k++, " ");
}

void dup_song(song_t *d, song_t s)
{
	snprintf(d->name, MAX_NAME_LEN, s.name);
	snprintf(d->path, MAX_PATH_LEN, s.path);
	d->start=s.start;
	d->end=s.end;
	d->isdir=s.isdir;
}

int randomise_play_list(int listlen)
{
	int i=0;
	int s=0;
	FILE *fp=NULL;
	if(!(fp=fopen("/dev/random", "r")))
	{
		endwin();
		printf("Cannot open /dev/random for reading.\n");
		return -1;
	}
	s=getc(fp);
	fclose(fp);
	song_t rlist[listlen];
	for(i=0; i<listlen; i++)
		snprintf(rlist[i].path, MAX_PATH_LEN, "");
	srand(s);
	for(i=0; i<listlen; i++)
	{
		int j=0;
		do { j=rand()%listlen; } while(strlen(rlist[j].path));
		dup_song(&rlist[j], play_list[i]);
	}
	for(i=0; i<listlen; i++)
		dup_song(&play_list[i], rlist[i]);
	return 0;
}

int search_play_list(song_t search, int listlen)
{
	int i=0;
	for(i=0; i<listlen; i++)
	{
		if(strcmp(play_list[i].name, search.name)==0
			&& strcmp(play_list[i].path, search.path)==0)
				return i;
	}
	return -1;
}

int dir_search_play_list(song_t search, int listlen)
{
	int i=0;
	int c=0;
	for(i=0; i<listlen; i++)
	{
		if(strncmp(play_list[i].path, search.path, strlen(search.path))==0)
			c++;
	}
	return c;
}

int print_list(int view, int count, int listlen, int start,
					int row, int col, int sel)
{
	int a=0;
	int j=TOP_SPACE;
	int ctu=view?count:listlen;

	for(a=0; a<ctu; a++)
	{
		if(a>=start)
		{
			if(view)
			{
				if(dir_list[a].isdir) attron(COLOR_PAIR(2));
				else attron(COLOR_PAIR(1));
			}
			else
			{
				attron(A_BOLD);
		//		attron(COLOR_PAIR(3));
			}
			
			print_truncated(view?dir_list[a].name:play_list[a].name, j, col, selected==a?1:0);
			if(view)
			{
				if(dir_list[a].isdir) attroff(COLOR_PAIR(2));
				else attroff(COLOR_PAIR(1));
			}
			else
			{
				attroff(A_BOLD);
	//			attroff(COLOR_PAIR(3));
			}
			if(view)
			{
				int p=0;
				if(dir_list[a].isdir)
				{
					p=dir_search_play_list(dir_list[a], listlen);
					if(p)
					{
						attron(COLOR_PAIR(2));
						mvprintw(j, 0, "%3d", p);
						attron(A_BOLD);
						print_truncated(dir_list[a].name, j, col, selected==a?1:0);
						attroff(A_BOLD);
						attroff(COLOR_PAIR(2));
					}
					else
						mvprintw(j, 0, "   ");
				}
				else
				{
					p=search_play_list(dir_list[a], listlen);
					if(p>=0)
					{
						mvprintw(j, 0, "%3d", p+1);
				//		attron(COLOR_PAIR(3));
						attron(A_BOLD);
						print_truncated(dir_list[a].name, j, col, selected==a?1:0);
					//	attroff(COLOR_PAIR(3));
						attroff(A_BOLD);
					}
					else
						mvprintw(j, 0, "   ");
				}
			}
			else mvprintw(j, 0, "%3d", a+1);
			j++;
		}
	}
	while(j<=row-TOP_SPACE+1)
	{
		int k=0;
		while(k<col-1) mvprintw(j, k++, "   ");
		j++;
	}
	return 0;
}

void add_song_to_beginning(int listlen, int selected)
{
	int i=0;
	song_t tmp;

	dup_song(&tmp, play_list[0]);
	dup_song(&play_list[0], dir_list[selected]);

	for(i=0; i<listlen; i++)
	{
		if(strcmp(tmp.path, dir_list[selected].path))
			dup_song(&play_list[i], tmp);
	}
}

void add_song_to_list(int *listlen, int selected)
{
	if(search_play_list(dir_list[selected], *listlen)<0)
	{
		snprintf(play_list[*listlen].name, MAX_NAME_LEN,
					dir_list[selected].name);
		snprintf(play_list[*listlen].path, MAX_PATH_LEN,
					dir_list[selected].path);
		play_list[*listlen].start=dir_list[selected].start;
		play_list[*listlen].end=dir_list[selected].end;
		play_list[*listlen].isdir=0;
		(*listlen)++;
	}
}

void remove_song_from_list(int view, int *listlen, int selected)
{
	int j=0;
	int i=0;

	if(view) i=search_play_list(dir_list[selected], *listlen);
	else i=selected;

	if(i>=0 && i<*listlen)
	{
		for(j=i+1; j<*listlen; i++, j++)
		{
			dup_song(&play_list[i], play_list[j]);
		}
		(*listlen)--;
	}
}

void print_header(int view, char *dir_to_scan, int selected, int paused, int col)
{
	print_truncated("", 1, col, 0);
	if(view)
		print_truncated(dir_to_scan, 0, col, 0);
	else
	{
		char *cp=NULL;
		char header[MAX_PATH_LEN]="";
		snprintf(header, MAX_PATH_LEN, play_list[selected].path);
		if((cp=strrchr(header, '/'))) *cp='\0';

		print_truncated(header, 0, col, 0);
	}
	if(paused) print_truncated("paused", 0, col, 0);
	else if(playing)
	{
		int pos;
		int len=play_list[0].end-play_list[0].start;
		ap_get_frame(0, &pos);
		pos-=play_list[0].start;
		mvprintw(0, col-14, " %5d / %-5d", pos>0?pos:0, len>0?len:0);
	}
}

void rotate_list(int *listlen)
{
	int i=0;
	song_t tmp;

	if(looping) dup_song(&tmp, play_list[0]);
	for(i=0; i<(*listlen)-1; i++)
		dup_song(&play_list[i], play_list[i+1]);
	if(looping) dup_song(&play_list[i], tmp);
	if(!looping) (*listlen)--;
}

int update_player(int *listlen)
{
	if(!*listlen)
	{
		ap_stop(0);
		ap_clear_playlist(0);
		playing=0;
	}
	if(*listlen)
	{
		if(!playing)
		{
			playing=1;
			ap_add_path(0, play_list[0].path);
			ap_next(0);
			ap_play(0);
			ap_set_frame(0, play_list[0].start);
			dup_song(&current, play_list[0]);
		}
		else
		{
			int pos;
			ap_get_frame(0, &pos);
			if(strcmp(play_list[0].name, current.name)
				|| strcmp(play_list[0].path, current.path))
			{
				ap_stop(0);
				ap_clear_playlist(0);
				playing=0;
			}
			else if(pos>=play_list[0].end || !pos)
			{
				rotate_list(listlen);

				// only stop playing if the next track doesn't continue from the first one!
				if(strcmp(current.path, play_list[0].path)==0
				  && current.end==play_list[0].start)
				{
					dup_song(&current, play_list[0]);
				}
				else
				{
					ap_stop(0);
					ap_clear_playlist(0);
					playing=0;
				}
			}
		}
	}

	return 0;
}

void swap_songs(song_t *a, song_t *b)
{
	song_t tmp;
	dup_song(&tmp, *a);
	dup_song(a, *b);
	dup_song(b, tmp);
}

void printhfnc(char *string, int row, int col)
{
	mvprintw(row, 0, "   ");
	print_truncated(string, row, col, 0);
}

void print_help(int row, int col)
{
	int i=0;
	for(i=0; i<row; i++)
		printhfnc("", i, col);
	i=2;
	printhfnc("help", i++, col);
	printhfnc("----", i++, col);
	i+=2;
	printhfnc("         up/ k - move cursor up     a - add song to playlist", i++, col);
	printhfnc("      down / j - move cursor down   d - delete song from playlist", i++, col);
	printhfnc("      left / h - enter directory    t - toggle browse / playlist mode", i++, col);
	printhfnc("     right / l - leave directory    c - clear playlist", i++, col);
	printhfnc("   page up / b - page up            p - pause / unpause", i++, col);
	printhfnc(" page down / space - page down      g - randomise playlist", i++, col);
	printhfnc("             x - fast forward       n - move track up the playlist", i++, col);
	printhfnc("             z - rewind             m - move track down playlist", i++, col);
	printhfnc("             w - write playlist     y - move track to top of playlist", i++, col);
	printhfnc("             e - erase playlist     u - move track to bottom of playlist", i++, col);
	printhfnc("             s - skip current       o - toggle playlist looping on / off", i++, col);
	printhfnc("             r - restart current    ? - this help screen", i++, col);
	i+2;
	printhfnc("Press any key to exit help", ++i, col);
	
}

int save_playlist(const char *loc, int listlen)
{
	int i=0;
	mode_t mtu;
	FILE *fp=NULL;
	char buf[256]="";
	mode_t mask=umask(0755);

	mtu=umask(mask);
	snprintf(buf, sizeof(buf), "%s/%s", SAVE_DIR, loc);

	i=mkdir(buf, mtu);
	if(i && errno!=EEXIST)
	{
		fprintf(stderr, "Unable to mkdir(%s): %s\n", buf, strerror(errno));
		return -1;
	}
	snprintf(buf, sizeof(buf), "%s/%s/%s", SAVE_DIR, loc, CFG);
	if(!(fp=fopen(buf, "w")))
	{
		fprintf(stderr, "Unable to open '%s' for writing: %s\n", buf, strerror(errno));
		return -1;
	}
	for(i=0; i<listlen; i++)
		fprintf(fp, "%s:%05d:%05d:%s\n", play_list[i].path, play_list[i].start, play_list[i].end, play_list[i].name);
	if(fclose(fp))
	{
		fprintf(stderr, "Closing file '%s': %s\n", buf, strerror(errno));
		return -1;
	}
	return 0;
}

int remove_playlist(const char *path)
{
	char buf[MAX_PATH_LEN]="";

	snprintf(buf, sizeof(buf), "%s/%s", path, CFG);
	if(unlink(buf)) return -1;
	if(rmdir(path)) return -1;
	return 0;
}

int main(int argc, char *argv[])
{
	int row;
	int col;
	int ch;
	int option;
	int i=0;
	int count=0;
	int startpos=0;
	int listlen=0;
	int view=1;
	int paused=0;
	int help=0;
	int quit=0;
	int save=0;
	int erase=0;
	char dir_to_scan[MAX_PATH_LEN]="";
	char *firstdir=NULL;
	char *origdir=NULL;
	char savebuf[MAX_SAVE_LEN]="";
	char buf[128]="";
	depth_t dpth[MAX_DEPTH];
	depth_t dpthl;
	int dpthd=0;

	while((option=getopt(argc, argv, "h?"))!=-1)
	{
		switch(option)
		{
			case 'h':
			case '?':
				fprintf(stdout, "%s\n", USAGE);
				return 1;
			default:
				fprintf(stderr, "flag '-%c %s' unknown", option, optarg);
			break;
			
		}
	}

	if(argc<2 || optind>=argc)
	{
		fprintf(stdout, USAGE);
		return 1;
	}

	firstdir=argv[argc-1];
	if(strncmp(firstdir, "./", strlen("./"))==0) firstdir+=strlen("./");
	printf("%s\n", firstdir);
	if(firstdir[strlen(firstdir)-1]=='/' &&
		firstdir[strlen(firstdir)-1]!=firstdir[0]) firstdir[strlen(firstdir)-1]='\0';

	if(strlen(firstdir) && chdir(firstdir))
	{
		fprintf(stderr, "Unable to chdir to '%s'\n", firstdir);
		return 1;
	}
	if(get_fullpath(firstdir, dir_to_scan, MAX_PATH_LEN))
	{
		fprintf(stderr, "Unable to get full directory path.\n");
		return 1;
	}
	if(!getcwd(dir_to_scan, sizeof(dir_to_scan)))
	{
		fprintf(stderr, "Could not get current working directory in main\n");
		return -1;
	}
	if(!(firstdir=strdup(dir_to_scan)))
	{
		fprintf(stderr, "Unable to strdup dir_to_scan '%s'.\n", dir_to_scan);
		return 1;
	}

	for(i=0; i<MAX_DEPTH; i++)
	{
		dpth[i].startpos=0;
		dpth[i].selected=0;
	}
	dpthl.startpos=0;
	dpthl.selected=0;

	initscr();
	start_color();
	init_pair(1, COLOR_WHITE, COLOR_BLACK);
	init_pair(2, COLOR_WHITE, COLOR_BLACK);
	init_pair(3, COLOR_WHITE, COLOR_BLACK);
	raw();
	keypad(stdscr, TRUE);
	noecho();
	curs_set(0);

	if(get_dir(dir_to_scan, &count)) return 1;
	getmaxyx(stdscr, row, col);

	print_header(view, dir_to_scan, selected, paused, col);
	print_list(view, count, listlen, startpos, row, col, selected);
	halfdelay(3);
	while((ch=getch()) && ap_session_running(0))
	{
		refresh();
		getmaxyx(stdscr, row, col);
		if(update_player(&listlen))
		{
			endwin();
			return 1;
		}
		if(ch==ERR)
		{
		}
		else
		{
			int pos;
			char *cp=NULL;
			int pagelen=row-TOP_SPACE;
			int ctu=view?count:listlen;

			if(help) help=0;
			else if(quit)
			{
				switch(ch)
				{
					case 'y':
					case 'Y':
					case 10:
					case 13:
						endwin();
						return 1;
						break;
					case 'n':
					case 'N':
						quit=0;
						break;
					default:
						break;
				}
			}
			else if(erase)
			{
				switch(ch)
				{
					case 'Y':
					case 'y':
					case 10:
					case 13:
						if(remove_playlist(dir_list[selected].path))
						{
							endwin();
							return 1;
						}
						count=0;
						if(get_dir(dir_to_scan, &count))
						{
							endwin();
							return 1;
						}
						erase=0;
						break;
					case 'n':
					case 'N':
						erase=0;
						break;
					default:
						break;
				}
			}
			else if(save)
			{
				char tmp[MAX_SAVE_LEN]="";

				if(ch==10 || ch==13)
				{
					if(strlen(savebuf)>0)
					{
						if(save_playlist(savebuf, listlen))
						{
							endwin();
							return 1;
						}
					}
					save=0;
				}
				else if(check_validchars(ch))
				{
					snprintf(tmp, sizeof(tmp), "%s", savebuf);
					snprintf(savebuf, sizeof(savebuf), "%s%c", tmp, ch);
				}
				else if(ch==KEY_BACKSPACE)
				{
					if(strlen(savebuf)>0)
						savebuf[strlen(savebuf)-1]='\0';
					else
						save=0;
				}
			}
			else switch(ch)
			{
				case 'w':
				case 'W':
					if(listlen) save=1;
					break;
				case 'q':
				case 'Q':
					quit=1;
					break;
				case 'k':
				case 'K':
				case KEY_UP:
					if(selected>0)
					{
						if(selected<=startpos) startpos--;
						selected--;
					}
					break;
				case 'j':
				case 'J':
				case KEY_DOWN:
					if(selected+1<ctu)
					{
						selected++;
						if(selected>pagelen-1) startpos++;
					}
					break;
				case ' ':
				case KEY_NPAGE:
					if(selected+(pagelen)<ctu)
					{
						selected+=(pagelen);
						if(selected>pagelen-1) startpos+=(pagelen);
					}
					else
					{
						startpos=pagelen;
						selected=ctu-1;
					}
					if(pagelen>count) startpos=0;
					break;
				case 'b':
				case 'B':
				case KEY_PPAGE:
					if(selected-(pagelen)>0)
					{
						selected-=(pagelen);
						if(selected<TOP_SPACE) startpos-=(pagelen);
					}
					else selected=0;
					if(selected<startpos) startpos=selected;
					if(pagelen>ctu) startpos=0;
					break;
				case 'h':
				case 'H':
				case KEY_LEFT:
					if(!view || strcmp(dir_to_scan, firstdir)==0)
						break;
					if((cp=strrchr(dir_to_scan, '/')) && cp!=dir_to_scan) *cp='\0';
					else snprintf(dir_to_scan, MAX_PATH_LEN, "/");
					if(chdir("../"))
					{
						endwin();
						fprintf(stderr, "Unable to chdir to '%s'\n", dir_to_scan);
						return 1;
					}
					dpth[dpthd].startpos=startpos;
					dpth[dpthd].selected=selected;
					dpthd--;
					count=0;
					startpos=dpth[dpthd].startpos;
					selected=dpth[dpthd].selected;
					if(get_dir(dir_to_scan, &count))
					{
						endwin();
						return 1;
					}
					break;
				case 'l':
				case 'L':
				case KEY_RIGHT:
					if(view && count && dir_list[selected].isdir)
					{
						if(chdir(dir_list[selected].path))
						{
							endwin();
							fprintf(stderr, "Unable to chdir to '%s'\n", dir_to_scan);
							return 1;
						}
						dpth[dpthd].startpos=startpos;
						dpth[dpthd].selected=selected;
						dpthd++;
						count=0;
						snprintf(dir_to_scan, MAX_PATH_LEN, dir_list[selected].path);
						if(get_dir(dir_to_scan, &count))
						{
							endwin();
							return 1;
						}
						startpos=dpth[dpthd].startpos;
						selected=dpth[dpthd].selected;
						if(selected>=listlen-1)
						{
							selected=0;
							startpos=0;
						}
					}
					break;
				case 't':
				case 'T':
					if(view)
					{
						if(listlen)
						{
							view=0;
							dpth[dpthd].startpos=startpos;
							dpth[dpthd].selected=selected;
							selected=dpthl.selected;
							startpos=dpthl.startpos;
							if(selected>=listlen) selected=listlen-1;
						}
					}
					else
					{
						view=1;
						dpthl.selected=selected;
						dpthl.startpos=startpos;
						selected=dpth[dpthd].selected;
						startpos=dpth[dpthd].startpos;
					}
					break;
				case 'a':
				case 'A':
					if(listlen<MAX_LIST_LEN-1 && view)
					{
						if(!dir_list[selected].isdir)
						{
							add_song_to_list(&listlen, selected);
							if(selected+1<ctu)
							{
								selected++;
								if(selected>pagelen-1) startpos++;
							}
						}
						else if(count)
						{
							int i=0;
							char *cp=NULL;
							if(chdir(dir_list[selected].path))
							{
								endwin();
								fprintf(stderr, "Unable to enter directory '%s'\n", dir_list[selected].path);
								return 1;
							}
							count=0;
							snprintf(dir_to_scan, MAX_PATH_LEN, dir_list[selected].path);
							if(get_dir(dir_to_scan, &count))
							{
								endwin();
								return 1;
							}
							for(i=0; i<count; i++)
							{
								if(!dir_list[i].isdir) add_song_to_list(&listlen, i);
							}
							if((cp=strrchr(dir_to_scan, '/')) && cp!=dir_to_scan) *cp='\0';
							else snprintf(dir_to_scan, MAX_PATH_LEN, "/");
							if(chdir("../"))
							{
								endwin();
								fprintf(stderr, "Unable to leave directory '%s'\n", dir_list[selected].path);
								return 1;
							}
							count=0;
							if(get_dir(dir_to_scan, &count))
							{
								endwin();
								return 1;
							}
						}
					}
					break;
				case 'd':
				case 'D':
					if(!view || !dir_list[selected].isdir)
					{
						remove_song_from_list(view, &listlen, selected);
						if(!view && selected>listlen-1)
						{
							selected--;
						}
						// if nothing in list, revert to browse view.
						if(!view && listlen==0)
						{
							view=1;
							selected=0;
							startpos=0;
						}
					}
					break;
				case 'p':
				case 'P':
					if(paused)
					{
						ap_unpause(0);
						paused=0;
					}
					else
					{
						ap_pause(0);
						paused=1;
					}
					break;
				case 'z':
				case 'Z':
					ap_get_frame(0, &pos);
					if(pos>current.start+200)
					{
						ap_stop(0);
						ap_clear_playlist(0);
						ap_add_path(0, current.path);
						ap_next(0);
						ap_play(0);
						ap_set_frame(0, pos-200);
					}
					break;
				case 'x':
				case 'X':
					ap_get_frame(0, &pos);
					if(pos>current.start-200)
					{
						ap_stop(0);
						ap_clear_playlist(0);
						ap_add_path(0, current.path);
						ap_next(0);
						ap_play(0);
						ap_set_frame(0, pos+200);
					}
					break;
				case 'n':
				case 'N':
					if(!view)
					{
						if(selected>0)
						{
							swap_songs(&play_list[selected], &play_list[selected-1]);
							selected--;
						}
					}
					else
					{
						int p=search_play_list(dir_list[selected], listlen);
						if(p>0)
						{
							swap_songs(&play_list[p], &play_list[p-1]);
						}
					}
					break;
				case 'm':
				case 'M':
					if(!view)
					{
						if(selected<listlen-1)
						{
							swap_songs(&play_list[selected], &play_list[selected+1]);
							selected++;
						}
					}
					else
					{
						int p=search_play_list(dir_list[selected], listlen);
						if(p>=0 && p<listlen-1)
						{
							swap_songs(&play_list[p], &play_list[p+1]);
						}
					}
					break;
				case 'c':
				case 'C':
					listlen=0;
					if(!view)
					{
						view=1;
						selected=0;
						startpos=0;
					}
					break;
				case '?':
					help=1;
					break;
				case 's':
				case 'S':
					if(listlen)
						rotate_list(&listlen);
					break;
				case 'r':
				case 'R':
					if(playing)
						ap_set_frame(0, current.start);
					break;
				case 'e':
				case 'E':
					// only erase playlists in SAVE_DIR
					if(view
						&& strncmp(dir_list[selected].path, SAVE_DIR, strlen(SAVE_DIR))==0)
					{
						erase=1;
					}
					break;
				case 'o':
				case 'O':
					if(looping)
					{
						print_truncated("playlist looping off", 0, col, 0);
						looping=0;
					}
					else
					{
						print_truncated("playlist looping on", 0, col, 0);
						looping=1;
					}
					refresh();
					sleep(1);
					break;
				case 'g':
				case 'G':
					if(randomise_play_list(listlen))
						return 1;
					break;
				case 'y':
				case 'Y':
					if(!view || !dir_list[selected].isdir)
					{
						add_song_to_beginning(listlen, selected);
					}
					break;
				case 'u':
				case 'U':
					if(!view || !dir_list[selected].isdir)
					{
						remove_song_from_list(view, &listlen, selected);
						add_song_to_list(&listlen, selected);
					}
					break;
			}
			if(selected>startpos+pagelen) selected=pagelen;
			usleep(20000);
			flushinp();
		}
		if(help) print_help(row, col);
		else if(quit)
			print_truncated("quit (yes/no)", 0, col, 0);
		else if(erase)
		{
			snprintf(buf, sizeof(buf), "erase playlist '%s' (yes/no)", dir_list[selected].name);
			print_truncated(buf, 0, col, 0);
		}
		else if(save)
		{
			snprintf(buf, sizeof(buf), "save playlist to: %s", savebuf);
			print_truncated(buf, 0, col, 0);
		}
		else
		{
			print_header(view, dir_to_scan, selected, paused, col);
			print_list(view, count, listlen, startpos, row, col, selected);
		}
	}
	endwin();
	if(!(ap_session_running(0)))
	{
		fprintf(stderr, "No alsaplayer session running!\n");
		return 1;
	}

	return 0;
}
